/** SelectByString
  *     A version AviSynth's Select function that operates on strings.
  *
  * Last modified: 2004-05-30
  *
  * Copyright (c) 2004 James D. Lin
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
  * met:
  *
  * * Redistributions of source code must retain the above copyright
  *   notice, this list of conditions and the following disclaimer.
  *
  * * Redistributions in binary form must reproduce the above copyright
  *   notice, this list of conditions and the following disclaimer in the
  *   documentation and/or other materials provided with the distribution.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
  * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
  * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
  * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */

#include "windows.h"
#include "avisynth.h"



/** strEqual
  *
  *     Determines if <s> and <t> are equivalent strings.
  *
  *     Unlike strcmp, allows case-insensitive comparisons.
  *
  * PARAMETERS:
  *     IN s, t    - the strings to compare
  *     ignoreCase - pass true to perform a case-insensitive comparison
  *
  * PRE:
  *     <s> and <t> are null-terminated ASCII strings
  *
  * RETURNS:
  *     true if <s> and <t> are equal;
  *     false otherwise
  */
static bool strEqual(const char* s, const char* t, bool ignoreCase)
{
    if (s == t) { return true; }

    while (true)
    {
        if (*s == '\0' && *t == '\0')
        {
            return true;
        }
        else if (*s == *t || (ignoreCase && tolower(*s) == tolower(*t)))
        {
            ++s;
            ++t;
        }
        else
        {
            return false;
        }
    }
}


/** selectByString
  *
  *     Callback passed to the scripting environment.
  *
  * PARAMETERS:
  *     args         - the array of arguments passed to this filter from
  *                      the script
  *     IN userDataP - the exposed name of this function
  *     IN/OUT envP  - pointer to the AviSynth scripting environment
  *
  * RETURNS:
  *     the selected AviSynth value
  */
static AVSValue __cdecl selectByString(AVSValue args, void* userDataP, IScriptEnvironment* envP)
{
    const char* nameP = static_cast<char*>(userDataP);
    const AVSValue* vP;

    const char* keyStr = args[0].AsString();
    const AVSValue& items = args[1];
    const AVSValue& elseItem = args[2];

    int n = items.ArraySize();

    if (n % 2 != 0)
    {
        envP->ThrowError("%s: keys and values must occur in pairs", nameP);
    }
    else
    {
        int i;
        const char* s;

        // do a simple linear search
        for (i = 0; i < n; i += 2)
        {
            if (!items[i].IsString())
            {
                // report key indices as 1-based
                envP->ThrowError("%s: key #%d is not a string", nameP, i / 2 + 1);
                break;
            }

            s = items[i].AsString();
            if (s != NULL && strEqual(keyStr, s, true))
            {
                vP = &items[i + 1];
                break;
            }
        }
        if (i == n)
        {
            // value not found
            if (elseItem.Defined())
            {
                vP = &elseItem;
            }
            else
            {
                envP->ThrowError("%s: key \"%s\" not found", nameP, keyStr);
            }
        }
    }
    return *vP;
}


/** selectByStringEval
  *
  *     Callback passed to the scripting environment.
  *
  * PARAMETERS:
  *     args         - the array of arguments passed to this filter from
  *                      the script
  *     IN userDataP - the exposed name of this function
  *     IN/OUT envP  - pointer to the AviSynth scripting environment
  *
  * RETURNS:
  *     the evaluated value of the selected AviSynth string
  */
static AVSValue __cdecl selectByStringEval(AVSValue args, void* userDataP, IScriptEnvironment* envP)
{
    const char* nameP = static_cast<char*>(userDataP);
    AVSValue v = selectByString(args, userDataP, envP);
    try
    {
        v = envP->Invoke("Eval", v);
    }
    catch (const IScriptEnvironment::NotFound&)
    {
        envP->ThrowError("%s: Could not invoke Eval", nameP);
    }
    return v;
}


/** AddPluginFunction
  *
  *     Helper function to add a function to the AviSynth scripting
  *     environment.
  *
  * PARAMETERS:
  *     IN nameP    - the exposed name of the added function
  *     IN paramsP  - the format string for the parameters to exposed
  *                     function
  *     apply       - pointer to the function to add
  *     IN/OUT envP - pointer to the AviSynth scripting environment
  *
  * SIDE EFFECTS:
  *     adds the specified filter to the AviSynth scripting environment
  */
static void AddPluginFunction(const char* nameP, const char* paramsP,
                              IScriptEnvironment::ApplyFunc apply, IScriptEnvironment* envP)
{
    envP->AddFunction(nameP, paramsP, apply, const_cast<char*>(nameP));
}


/** AvisynthPluginInit2
  *
  *     AviSynth 2.5x plug-in entry-point.
  *
  * PARAMETERS:
  *     IN/OUT envP - pointer to the AviSynth scripting environment
  *
  * RETURNS:
  *     the version string for this plug-in
  *
  * SIDE EFFECTS:
  *     adds functions to the AviSynth scripting environment
  */
extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* envP)
{
    AddPluginFunction("SelectByString", "s.+[else].", selectByString, envP);
    AddPluginFunction("SelectByStringEval", "ss+[else]s", selectByStringEval, envP);
    return "SelectByString v0.3.0 [" __DATE__ "]\nCopyright (c) 2004 James D. Lin";
}
